/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/getarch.c,v 1.1 1993/02/11 21:43:13 mcooper Exp mcooper $";
#endif

/*
 * Get architecture information
 */

#include "defs.h"

extern char		       *AppArchCmds[];
extern char		       *KernArchCmds[];
extern char		       *ArchFiles[];

#if	defined(HAVE_SYSINFO)
/*
 * Get kernel arch using sysinfo() system call.
 */
static char *GetKernArchFromSysinfo()
{
    static char			buff[BUFSIZ];

    if (buff[0])
	return(buff);

    if (sysinfo(SI_MACHINE, buff, sizeof(buff)) < 0)
	return((char *) NULL);

    return(buff);
}

/*
 * Get application arch using sysinfo() system call.
 */
static char *GetAppArchFromSysinfo()
{
    static char			buff[BUFSIZ];

    if (buff[0])
	return(buff);

    if (sysinfo(SI_ARCHITECTURE, buff, sizeof(buff)) < 0)
	return((char *) NULL);

    return(buff);
}
#endif	/* HAVE_SYSINFO */

/*
 * Get application architecture.
 */
extern char *GetAppArch()
{
    static char		       *arch = NULL;

    if (arch)
	return(arch);

#if	defined(ARCH_TYPE)
    /*
     * Use predefined name if present.
     */
    if (!arch)
	return(ARCH_TYPE);
#endif	/* ARCH_TYPE */

#if	defined(HAVE_SYSINFO)
    /*
     * Use sysinfo() method
     */
    if (arch = GetAppArchFromSysinfo())
	return(arch);
#endif	/* HAVE_SYSINFO */

#if	defined(HAVE_AARCHNAME)
    /*
     * Use OS specific method
     */
    if (arch = (char *) GetAppArchName())
	return(arch);
#endif	/* HAVE_AARCHNAME */

    /*
     * Try running App Arch commands
     */
    if (arch = RunCmds(AppArchCmds))
	return(arch);

    /*
     * Try testing Architecture files
     */
    if (arch = RunTestFiles(ArchFiles))
	return(arch);

    return(arch);
}

/*
 * Get kernel architecture
 */
extern char *GetKernArch()
{
    char 		       *arch = NULL;

#if	defined(KARCH_TYPE)
    /*
     * If there's a predefined KArch name, use it.
     */
    if (!arch)
	return(KARCH_TYPE);
#endif	/* KARCH_TYPE */

#if	defined(HAVE_SYSINFO)
    if (arch = GetKernArchFromSysinfo())
	return(arch);
#endif	/* HAVE_SYSINFO */

#if	defined(HAVE_KARCHNAME)
    /*
     * Try OS specific method
     */
    if (arch = GetKernArchName())
	return(arch);
#endif	/* HAVE_KARCHNAME */

    /*
     * Try running the KArch test commands
     */
    if (arch = RunCmds(KernArchCmds))
	return(arch);

    /*
     * We're desperate, so try using the Application Architecture
     */
    if (arch = GetAppArch())
	return(arch);

    return((char *) NULL);
}

