/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/getosver.c,v 1.2 1993/06/08 20:15:09 mcooper Exp mcooper $";
#endif

/*
 * Get OS version information
 */

#include "defs.h"

#if	defined(HAVE_UNAME)
/*
 * Get OS version number using uname()
 */
static char *GetOSVersionFromUname()
{
    static char			Buf[BUFSIZ];
    struct utsname 		un;

    if (uname(&un) != 0)
	return((char *) NULL);

    /*
     * Vendors don't all do the same thing for storing
     * version numbers via uname().
     */
#if	defined(UNAME_REL_VER_COMB)
    (void) sprintf(Buf, "%s.%s", un.version, un.release);
#else
    (void) sprintf(Buf, "%s", un.release);
#endif 	/* UNAME_REL_VER_COMB */

    return(Buf);
}
#endif	/* HAVE_UNAME */

#if	defined(OSVERS_FROM_KERNVER)
/*
 * Get OS version by reading an a specific argument out of
 * the kernel version string.
 */
extern char *
GetOSVersionFromKernVer()
{
    static char			Buf[BUFSIZ];
    register char	       *cp;
    register int		i;

    if (!(cp = GetKernelVersion()))
	return((char *) NULL);

    (void) strcpy(Buf, cp);
    for (cp = strtok(Buf, " "), i = 0; cp && i != OSVERS_FROM_KERNVER-1; 
	 cp = strtok((char *)NULL, " "), ++i);
    (void) strcpy(Buf, cp);
    if ((cp = index(Buf, ':')) != NULL)
	*cp = C_NULL;

    return(Buf);
}
#endif	/* OSVERS_FROM_KERNVER */

#if	defined(HAVE_SYSINFO)
/*
 * Get OS version using sysinfo() system call.
 */
static char *GetOSVersionFromSysinfo()
{
    static char			buff[BUFSIZ];

    if (buff[0])
	return(buff);

    if (sysinfo(SI_RELEASE, buff, sizeof(buff)) < 0)
	return((char *) NULL);

    return(buff);
}
#endif	/* HAVE_SYSINFO */

/*
 * Get Operating System version
 */
extern char *GetOSVersion()
{
    static char		       *osver = NULL;
    char 		       *Str;
    char 		       *p;

    if (osver)
	return(osver);

#if	defined(OS_VERSION)
    if (!osver)
	osver = OS_VERSION;
#endif	/* OS_VERSION */

#if	defined(HAVE_SYSINFO)
    if (!osver)
	osver = GetOSVersionFromSysinfo();
#endif	/* HAVE_SYSINFO */

#if	defined(HAVE_UNAME)
    if (!osver)
	osver = GetOSVersionFromUname();
#endif	/* HAVE_UNAME */

    if (!osver)
	osver = GetOSVersionStr();

    if (osver)
	/*
	 * Zap "*-PL*".
	 */
	if (*osver && ((p = rindex(osver, '-')) != NULL) && 
	    (strncmp(p, "-PL", 3) == 0))
	    *p = C_NULL;

    return(osver);
}
