/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */
#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/kvm.c,v 1.19 1993/06/08 21:00:52 mcooper Exp mcooper $";
#endif

/*
 * Frontend functions for kvm_*() functions
 *
 * It is assumed we HAVE_NLIST if we HAVE_KVM.
 */

#include <stdio.h>
#include "os.h"

#if	defined(HAVE_KVM)

#include "defs.h"
#include <fcntl.h>
#if	defined(HAVE_NLIST)
#	include <nlist.h>
#endif	/* HAVE_NLIST */

/*
 * Perform a kvm_close().  Really just hear to be compatible.
 */
extern void KVMclose(kd)
    kvm_t 		       *kd;
{
    if (kd)
	(void) kvm_close(kd);
}

/*
 * Perform a kvm_open().
 */
extern kvm_t *KVMopen()
{
    kvm_t 		       *kd = NULL;
    extern char 	       *ProgramName;

    if ((kd = kvm_open((char *)NULL, (char *)NULL, (char *)NULL, O_RDONLY,
		       ProgramName)) == NULL) {
	if (Debug) Error("kvm_open failed: %s.", SYSERR);
	return((kvm_t *) NULL);
    }

    return(kd);
}

#if	defined(HAVE_NLIST)

#if	defined(__MACH__)
#define		NL_NAME(p)	p->n_un.n_name
#else
#define		NL_NAME(p)	p->n_name
#endif	/* __MACH__ */

/*
 * Perform an nlist on "kd" looking for "Symbol".
 * If NlistPtr is not NULL, use it for storage.
 * Return the a pointer to the found nlist structure.
 */
extern struct nlist *KVMnlist(kd, Symbol, NlistPtr)
    kvm_t		       *kd;
    char		       *Symbol;
    struct nlist	       *NlistPtr;
{
    static struct nlist		nlistbuf;
    struct nlist	       *nlptr;

    if (NlistPtr)
	nlptr = NlistPtr;
    else {
	BZERO((char *) &nlistbuf, sizeof(nlistbuf));
	nlptr = &nlistbuf;
    }

#if	defined(COFF) || defined(ELF)
    /*
     * Skip over initial '_'
     */
    if (*Symbol == '_')
	NL_NAME(nlptr) = Symbol + 1;
    else
#endif
	NL_NAME(nlptr) = Symbol;

    if (kvm_nlist(kd, nlptr) != 0) {
	if (Debug) Error("kvm_nlist name \"%s\" failed: %s.", 
			 NL_NAME(nlptr), SYSERR);
	KVMclose(kd);
	return((struct nlist *) NULL);
    }

    return(nlptr);
}
#endif	/* HAVE_NLIST */

/*
 * Perform a kvm_read().
 */
extern int KVMread(kd, Addr, Buf, NumBytes)
    kvm_t 		       *kd;
    u_long	 		Addr;
    char 		       *Buf;
    unsigned 		        NumBytes;
{
    int				Count;

    if (!kd)
	return(-1);

    if ((Count = kvm_read(kd, Addr, Buf, NumBytes)) != NumBytes) {
	if (Debug) Error("kvm_read failed (expected %d, got %d): %s.", 
			 NumBytes, Count, SYSERR);
	return(-1);
    }

    return(0);
}

/*
 * Check to see if PtrNL is valid.
 */
extern int _CheckNlist(PtrNL)
    struct nlist	       *PtrNL;
{
    /*
     * Should use n_type, but that's not set
     * correctly on some OS's.
     */
    if (!PtrNL || !PtrNL->n_value) {
	if (Debug) Error("Kernel symbol \"%s\" not found.", 
			 GetNlNamePtr(PtrNL));
	return(-1);
    }

    return(0);
}

#endif /* HAVE_KVM */
