/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/netif.c,v 1.25 1993/06/09 18:12:13 mcooper Exp mcooper $";
#endif


/*
 * Portions of code found in this file are based on the 4.3BSD 
 * netstat(8) program.
 */

/*
 * Network Interface routines
 */

#include <stdio.h>
#include "os.h"

#include <fcntl.h>
#include <nlist.h>
#include <sys/types.h>
#include <sys/socket.h>
#if	defined(NEED_SOCKIO)
#include <sys/sockio.h>
#endif	/* NEED_SOCKIO */
#include <sys/param.h>
#include <sys/errno.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/if_ether.h>
#include <netdb.h>

#include "defs.h"

#if	GETNETIF_TYPE == GETNETIF_IFNET
/*
 * Network Interface symbol
 */
char 				NetifSYM[] = "_ifnet";
#endif	/* GETNETIF_IFNET */

/*
 * Interface Address union
 */
union {
    struct ifaddr 	ifaddr;
    struct in_ifaddr 	in_ifaddr;
} 				ifaddress;

/*
 * Create a DEVICE for a network interface.
 */
static DEVICE *CreateNetif(FullName, IfNet, DevData, DevDataTab)
    char		       *FullName;
    struct ifnet               *IfNet;
    DEVDATA		       *DevData;
    DEVDATATAB		       *DevDataTab;
{
    DEVICE		       *dev;

    dev = NewDevice(NULL);

    dev->dv_name 	= strdup(FullName);
    dev->dv_type 	= DT_NETIF;
    dev->dv_nodeid 	= DevData->dd_nodeid;
    dev->dv_unit 	= IfNet->if_unit;
    dev->dv_master 	= MkMasterFromDevData(DevData);

#if	defined(HAVE_IF_VERSION)
    if (IfNet->if_version && IfNet->if_version[0])
	dev->dv_model 	= strdup(IfNet->if_version);
    else
#endif	/* HAVE_IF_VERSION */
	dev->dv_model 	= DevDataTab->ddt_model;

    dev->dv_desc 	= DevDataTab->ddt_desc;

    return(dev);
}

/*
 * Return the netent of the network whose address is given.
 * The address is assumed to be that of a net or subnet, not a host.
 */
static struct netent *GetNet(inaddr, mask)
    u_long 			inaddr;
    u_long 			mask;
{
    u_long 			net;
    register u_long 		i, netaddr;
    int 			subnetshift;
    static struct in_addr 	in_addr;

    if (in_addr.s_addr = ntohl(inaddr)) {
	i = in_addr.s_addr;
	if (mask == 0) {
	    if (IN_CLASSA(i)) {
		mask = IN_CLASSA_NET;
		subnetshift = 8;
	    } else if (IN_CLASSB(i)) {
		mask = IN_CLASSB_NET;
		subnetshift = 8;
	    } else {
		mask = IN_CLASSC_NET;
		subnetshift = 4;
	    }
	    /*
	     * If there are more bits than the standard mask
	     * would suggest, subnets must be in use.
	     * Guess at the subnet mask, assuming reasonable
	     * width subnet fields.
	     */
	    while (in_addr.s_addr &~ mask)
		mask = (long)mask >> subnetshift;
	}
	net = in_addr.s_addr & mask;
	while ((mask & 1) == 0)
	    mask >>= 1, net >>= 1;
	netaddr = net;
    } else {
	netaddr = inaddr;
    }

    return(getnetbyaddr(netaddr, AF_INET));
}

/*
 * Get NETIF for an Internet address
 */
extern NETIF *GetNetifINET(aftab, hostaddr, maskaddr)
    AFTAB		       *aftab;
    struct sockaddr_in	       *hostaddr;
    struct sockaddr_in	       *maskaddr;
{
    struct in_addr		in_addr;
    struct netent 	       *np;
    struct hostent 	       *hp;
    struct sockaddr_in 	       *sin;
    char		       *inet_ntoa();
    NETIF		       *ni;

    ni = NewNetif(NULL);

    if (hostaddr && maskaddr) {
	np = GetNet(htonl(hostaddr->sin_addr.s_addr), 
		    htonl(maskaddr->sin_addr.s_addr));
	in_addr.s_addr = ntohl(hostaddr->sin_addr.s_addr & 
			       maskaddr->sin_addr.s_addr);
	ni->ni_netaddr = strdup(inet_ntoa(in_addr));
	sin = hostaddr;
    } else {
	np = GetNet(htonl(ifaddress.in_ifaddr.ia_subnet), 
		    ifaddress.in_ifaddr.ia_subnetmask);
	in_addr.s_addr = ntohl(ifaddress.in_ifaddr.ia_subnet);
	ni->ni_netaddr = strdup(inet_ntoa(in_addr));
	sin = (struct sockaddr_in *) &ifaddress.in_ifaddr.ia_addr;
    }

    ni->ni_hostaddr = strdup(inet_ntoa(sin->sin_addr));
    hp = gethostbyaddr((char *) &(sin->sin_addr),
		       sizeof(struct in_addr), AF_INET);

    if (hp)
	ni->ni_hostname = strdup(hp->h_name);

    if (np)
	ni->ni_netname = strdup(np->n_name);
	    
    if (aftab)
	ni->ni_typename = aftab->af_name;

    return(ni);
}

/*
 * Get NETIF for an unknown address type
 */
extern NETIF *GetNetifUnknown(aftab)
    AFTAB		       *aftab;
{
    NETIF		       *ni = NULL;

    ni = NewNetif(NULL);
    ni->ni_hostaddr = "<unknown>";
    if (aftab)
	ni->ni_typename = aftab->af_name;

    return(ni);
}

/*
 * Get an Address Family table entry
 */
static AFTAB *GetAFTab(type)
    int				type;
{
    extern AFTAB		AFTab[];
    register int		i;

    for (i = 0; AFTab[i].af_name; ++i)
	if (AFTab[i].af_type == type)
	    return(&AFTab[i]);

    return((AFTAB *) NULL);
}

#if	GETNETIF_TYPE == GETNETIF_IFNET
/*
 * Get a linked list of NETIF's for each address starting at 'startaddr'.
 */
static NETIF *GetNetifAddrs(kd, startaddr, FullName, Device)
    kvm_t		       *kd;
    off_t			startaddr;
    char		       *FullName;
    DEVICE		       *Device;
{
    u_long			addr;
    NETIF		       *base = NULL;
    register NETIF	       *ni, *pni;
    AFTAB		       *paftab;

    for (addr = startaddr; addr; addr = (u_long) ifaddress.ifaddr.ifa_next) {
	/*
	 * Read the ifaddr structure from kernel space
	 */
	if (KVMread(kd, addr, (char *) &ifaddress, sizeof(ifaddress))) {
	    Error("cannot read if address");
	    continue;
	}

	/*
	 * Now get and call the Address Family specific routine
	 * to extract a NETIF.
	 */
	if (paftab = GetAFTab(ifaddress.ifaddr.ifa_addr.sa_family)) {
	    if (ni = (*paftab->af_getnetif)(paftab, NULL, NULL))
		SetMacInfo(FullName, ni, Device);
	} else {
	    if (Debug) Error("Address family %d is not defined.", 
			     ifaddress.ifaddr.ifa_addr.sa_family);
	    continue;
	}

	/*
	 * Add the new NETIF to the base of the linked list.
	 */
	if (base) {
	    for (pni = base; pni && pni->ni_nxt; pni = pni->ni_nxt);
	    pni->ni_nxt = ni;
	} else {
	    base = ni;
	}
    }

    return(base);
}

/*
 * Query/find network interface devices and add them to devicelist
 */
extern DEVICE *ProbeNetif(name, DevData, DevDataTab)
    char 		       *name;
    DEVDATA 		       *DevData;
    DEVDATATAB	 	       *DevDataTab;
{
    DEVICE 		       *dev = NULL;
    static struct ifnet         ifnet;
    static char 		ifname[16], FullName[17];
    static struct nlist	        nlBuff;
    struct nlist	       *nlptr;
    register char	       *p;
    u_long 		        ifnetaddr;
    kvm_t 		       *kd;

    if (Debug) printf("ProbeNetif() '%s'\n", name);

    if (!(kd = KVMopen()))
	return((DEVICE *) NULL);

    /*
     * XXX - The unused "nlBuff" is required to avoid a nextstep gcc bug.
     */
    if ((nlptr = KVMnlist(kd, NetifSYM, &nlBuff)) == NULL)
	return((DEVICE *) NULL);

    if (CheckNlist(nlptr))
	return((DEVICE *) NULL);

    /*
     * Read address of ifnet structure from kernel space
     */
    if (KVMread(kd, nlptr->n_value, (char *) &ifnetaddr, sizeof(ifnetaddr))) {
	if (Debug) Error("kvm_read ifnetaddr failed");
	KVMclose(kd);
	return((DEVICE *) NULL);
    }

    /*
     * Read and then check each ifnet entry we found.
     */
    for (; ifnetaddr; ifnetaddr = (off_t) ifnet.if_next) {
	/*
	 * Read the ifnet structure
	 */
	if (KVMread(kd, ifnetaddr, (char *)&ifnet, sizeof(ifnet))) {
	    if (Debug) Error("kvm_read ifnetaddr ifnet failed");
	    continue;
	}

	/*
	 * Read if_name from kernel space
	 */
	if (KVMread(kd, (u_long)ifnet.if_name, ifname, sizeof(ifname))) {
	    if (Debug) Error("kvm_read ifname failed");
	    continue;
	}

	/*
	 * Copy if_name to 'FullName' and add unit number
	 */
	(void) sprintf(FullName, "%s%d", ifname, ifnet.if_unit);

	/*
	 * Check to see if this is the interface we want.
	 */
	if (!EQ(FullName, name))
	    continue;

	/*
	 * Create and set device
	 */
	dev = CreateNetif(FullName, &ifnet, DevData, DevDataTab);

	/*
	 * Get and set address info
	 */
	if (ifnet.if_addrlist) {
	    NETIF 	       *ni;

	    if (ni = GetNetifAddrs(kd, (off_t) ifnet.if_addrlist, 
				   FullName, dev))
		dev->dv_devspec = (caddr_t *) ni;
	}
    }

    KVMclose(kd);

    return(dev);
}
#endif	/* GETNETIF_IFNET */

#if	GETNETIF_TYPE == GETNETIF_IFCONF
/*
 * Set network address data.
 */
static int SetNetifAddrs(ifname, sock, hostaddr, maskaddr)
    char		       *ifname;
    int				sock;
    struct sockaddr_in	       *hostaddr;
    struct sockaddr_in	       *maskaddr;
{
    struct ifreq		ifr;

    (void) strcpy(ifr.ifr_name, ifname);

    /*
     * Get address info
     */
    if (ioctl(sock, SIOCGIFADDR, (char *) &ifr) < 0) {
	if (Debug) Error("%s: ioctl SIOCGIFADDR failed: %s.", ifname, SYSERR);
	return(-1);
    }
    BCOPY((char *)&ifr.ifr_addr, (char *)hostaddr, sizeof(struct sockaddr_in));

    /*
     * Get the network mask
     */
    if (ioctl(sock, SIOCGIFNETMASK, (char *) &ifr) < 0) {
	if (Debug) Error("%s: ioctl SIOCGIFNETMASK failed: %s.", 
			 ifname, SYSERR);
	return(-1);
    }
    BCOPY((char *)&ifr.ifr_addr, (char *)maskaddr, sizeof(struct sockaddr_in));

    return(0);
}

/*
 * Query/find network interface devices and add them to devicelist
 */
extern DEVICE *ProbeNetif(name, DevData, DevDataTab)
    char 		       *name;
    DEVDATA 		       *DevData;
    DEVDATATAB	 	       *DevDataTab;
{
    DEVICE 		       *dev = NULL;
    NETIF	 	       *ni;
    struct ifreq	       *ifreq;
    static struct ifconf	ifconf;
    static struct sockaddr_in	hostaddr;
    static struct sockaddr_in	maskaddr;
    static char			reqbuf[BUFSIZ];
    static char			buff[BUFSIZ];
    static int			sock = -1;
    register int		n;
    AFTAB		       *paftab;

    if (Debug) printf("ProbeNetif() '%s'\n", name);

    if (sock < 0) {
	/*
	 * Get list of all interfaces
	 */
	ifconf.ifc_len = sizeof(reqbuf);
	ifconf.ifc_buf = reqbuf;

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	    if (Debug) Error("Cannot create socket: %s.", SYSERR);
	    return((DEVICE *)NULL);
	}

	if (ioctl(sock, SIOCGIFCONF, (char *) &ifconf) < 0) {
	    if (Debug) Error("%s: ioctl SIOCGIFCONF failed: %s.", name,SYSERR);
	    return((DEVICE *)NULL);
	}
    }

    /*
     * Iterate over all known interfaces
     */
    for (ifreq = ifconf.ifc_req, n = ifconf.ifc_len/sizeof(struct ifreq);
	 --n >= 0; ifreq++) {

	/*
	 * Check to see if this is the interface we want.
	 */
	if (!EQ(name, ifreq->ifr_name))
	    continue;

	/*
	 * Create and set device
	 */
	dev 		= NewDevice(NULL);
	dev->dv_name 	= strdup(ifreq->ifr_name);
	dev->dv_type 	= DT_NETIF;
	dev->dv_nodeid 	= DevData->dd_nodeid;
	dev->dv_unit 	= DevData->dd_devunit;
	dev->dv_master 	= MkMasterFromDevData(DevData);
	dev->dv_model 	= DevDataTab->ddt_model;
	dev->dv_desc 	= DevDataTab->ddt_desc;

	/*
	 * Set address info
	 */
	if (SetNetifAddrs(ifreq->ifr_name, sock, &hostaddr, &maskaddr) == 0) {
	    /*
	     * Now get and call the Address Family specific routine
	     * to extract a NETIF.
	     */
	    if (paftab = GetAFTab(hostaddr.sin_family)) {
		if (ni = (*paftab->af_getnetif)(paftab, &hostaddr, 
						&maskaddr)) {
		    SetMacInfo(name, ni, dev);
		    dev->dv_devspec = (caddr_t *) ni;
		}
	    } else
		if (Debug) Error("Address family %d is not defined.", 
				 hostaddr.sin_family);
	}

	return(dev);
    }

    return((DEVICE *)NULL);
}
#endif	/* GETNETIF_IFCONF */
