/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/obp.c,v 1.18 1993/06/08 21:03:29 mcooper Exp mcooper $";
#endif

/*
 * Open Boot PROM (OBP) routines
 */

#include "defs.h"
#include <nlist.h>

#if	defined(HAVE_OPENPROM)

#if SUNOS == 5
#	include <sys/openprom.h>
#	include <sys/ddi.h>
#	include <sys/sunddi.h>
#	include <sys/ddi_impldefs.h>
#	define DEVI_CHILD(dev)		dev->devi_child
#	define DEVI_SIBLING(dev)	dev->devi_sibling
#	define DEVI_UNIT(dev)		dev->devi_instance
#	define DEVI_NODEID(dev)		dev->devi_nodeid
#	define DEVI_EXISTS(dev)		(dev->devi_addr && dev->devi_ops)
#else
#	include <sun/openprom.h>
#	define DEVI_CHILD(dev)		dev->devi_slaves
#	define DEVI_SIBLING(dev)	dev->devi_next
#	define DEVI_UNIT(dev)		dev->devi_unit
#	define DEVI_NODEID(dev)		0 /* No such member */
#	define DEVI_EXISTS(dev)		dev->devi_driver
#endif	/* SUNOS == 5 */

#define MAXOBPDESCLIST		100
#define OBP_IS_BOOL(tp)		(tp->obpt_key[strlen(tp->obpt_key)-1] == '?' \
				 || tp->obpt_value[0] == C_NULL)
/*
 * OBPio_t is used to do I/O with openprom.
 */
typedef union {
    char			opio_buff[OPROMMAXPARAM];
    struct openpromio		opio_oprom;
} OBPio_t;

/*
 * OBPtable_t is used to store properties for OBP node.
 */
typedef struct _obptable {
    char		        obpt_key[OPROMMAXPARAM];
    char		        obpt_value[OPROMMAXPARAM];
    struct _obptable	       *obpt_next;
} OBPtable_t;

/*
 * Various declarations
 */
extern char		 	OpenPROMSYM[];
typedef struct dev_info		MYdev_info_t;
static int 			OBPKtraverse();
static kvm_t		       *kd = NULL;
static long			CpuClockFreq = 0;

/*
 * Get the root (top) node from the OBP.
 */
static MYdev_info_t *OBPKgetRoot()
{
    static struct nlist		nlistbuf;
    struct nlist	       *nlptr;
    static MYdev_info_t	 	Root, *PtrRoot;
    u_long 			Addr;

    if (!kd)
	if (!(kd = KVMopen()))
	    return((MYdev_info_t *) NULL);

    if ((nlptr = KVMnlist(kd, OpenPROMSYM, &nlistbuf)) == NULL)
	return((MYdev_info_t *) NULL);

    if (CheckNlist(nlptr))
	return((MYdev_info_t *) NULL);

    /*
     * Read pointer to "top_devinfo" from kernel
     */
    Addr = nlptr->n_value;
    if (KVMread(kd, Addr, (char *) &PtrRoot, sizeof(struct dev_info *))) {
	if (Debug) Error("Cannot read openprom devinfo pointer from kernel");
	return((MYdev_info_t *) NULL);
    }

    if (KVMread(kd, (u_long)PtrRoot, (char *)&Root, sizeof(MYdev_info_t))) {
	if (Debug) Error("Cannot read openprom devinfo root from kernel");
	return((MYdev_info_t *) NULL);
    }

    return(&Root);
}

/*
 * Build device tree by reading the Open Boot PROM from the kernel.
 */
static int OBPKbuild(TreePtr)
    DEVICE 		      **TreePtr;
{
    MYdev_info_t	       *Root;

    if (!(Root = OBPKgetRoot()))
	return(-1);

    return(OBPKtraverse(Root, NULL, TreePtr));
}

/*
 * Cleanup an OBP node name.
 */
extern char *OBPcleanName(Name, FullClean)
    char		       *Name;
    int				FullClean;
{
    register char	       *cp;
    register char	       *end;
    register int		len;
    static char			buff[BUFSIZ];
    char		       *shortman;

    if (!Name || !Name[0])
	return((char *) NULL);

    (void) strcpy(buff, Name);

    /*
     * Remove all trailing spaces
     */
    end = &buff[strlen(buff)-1];
    for (cp = end; cp > &buff[0] && *cp == ' '; --cp);
    if (cp != end && cp && *++cp == ' ')
	*cp = C_NULL;

    if (FullClean) {
	/*
	 * Change '-' to ' '
	 * Change '_' to '/' (for SunOS 5.x)
	 */
	for (cp = &buff[0]; cp && *cp; ++cp) {
	    if (*cp == '-')
		*cp = ' ';
	    else if (*cp == '_')
		*cp = '/';
	}

	/*
	 * Skip over initial wording up to ','.  e.g. "SUNW,4/25" -> "4/25".
	 */
	if (cp = index(buff, ','))
	    ++cp;
	else
	    cp = buff;
    } else
	cp = buff;

    /*
     * Remove manufacturer part of name.
     * If shortman is "Sun" and Name is "Sun 4/25",
     * change Name to "4/25".
     */
    if (shortman = GetManShort()) {
	len = strlen(shortman);
	if (EQN(cp, shortman, len) && strlen(cp) > len+2 && cp[len] == ' ') {
	    cp += strlen(shortman);
	    while (cp && *cp && *cp == ' ')
		++cp;
	}
    }

    return(strdup(cp));
}

/*
 * Check an OpenPROM device.
 */
static int OBPKcheckDevice(DevInfo, Parent, TreePtr)
    MYdev_info_t		*DevInfo;
    MYdev_info_t		*Parent;
    DEVICE 		       **TreePtr;
{
    static DEVDATA 		 DevData;
    DEVICE 			*Device;	

    /* Make sure devdata is clean */
    BZERO(&DevData, sizeof(DEVDATA));
    DevData.dd_devunit 		= -1;
    DevData.dd_slave 		= -1;
    DevData.dd_ctlrunit 	= -1;
    DevData.dd_devnum 		= -1;

    /* Set what we know */
    if (DevInfo && DEVI_NODEID(DevInfo) != -1)
	DevData.dd_nodeid = DEVI_NODEID(DevInfo);

    if (DevInfo && DevInfo->devi_name && DevInfo->devi_name[0]) {
	DevData.dd_devname = OBPcleanName(DevInfo->devi_name, FALSE);
	/*
	 * If we have a parent, we're not the ROOT node.
	 * Otherwise we are the ROOT node.
	 * If so, we set the unit number to -1
	 * to avoid a bug in SunOS 4.x.
	 */
	if (Parent)
	    DevData.dd_devunit = DEVI_UNIT(DevInfo);
	else
	    DEVI_UNIT(DevInfo) = -1;
    }
    if (Parent && Parent->devi_name && Parent->devi_name[0]) {
	DevData.dd_ctlrname = OBPcleanName(Parent->devi_name, FALSE);
	DevData.dd_ctlrunit = DEVI_UNIT(Parent);
    }
    /* 
     * OpenPROM nodes that have a driver ALWAYS exist.
     * Some nodes may exist, without a driver, however.
     */
    if (DEVI_EXISTS(DevInfo))
	DevData.dd_flags |= DD_IS_ALIVE;

#if	SUNOS == 5
    /*
     * Get device number
     */
    if (DevInfo->devi_minor) {
	static struct ddi_minor_data	minor;

	if (KVMread(kd, (u_long) DevInfo->devi_minor, (char *) &minor,
		    sizeof(struct ddi_minor_data)) == 0 &&
	    minor.type == DDM_MINOR)
	    DevData.dd_devnum = minor.ddm_dev;
    }
#endif	/* SUNOS == 5 */

    if (Debug) {
	printf("OPENPROM: Found \"%s\" Node %d (Unit %d Dev %d) ",
	       ARG(DevData.dd_devname), DevInfo->devi_nodeid,
	       DevData.dd_devunit, DevData.dd_devnum);
	printf("on \"%s\" (Unit %d)%s\n",
	       ARG(DevData.dd_ctlrname), DevData.dd_ctlrunit,
	       (DevData.dd_flags & DD_IS_ALIVE) ? " [ALIVE]" : "");
    }

    /* Probe and add device */
    if (TreePtr && (Device = (DEVICE *) ProbeDevice(&DevData, TreePtr)))
	AddDevice(Device, TreePtr);
}

/*
 * Recursively traverse and descend the OpenPROM devinfo tree.
 */
static int OBPKtraverse(DevPtr, Parent, TreePtr)
    MYdev_info_t		*DevPtr;
    MYdev_info_t		*Parent;
    DEVICE 		       **TreePtr;
{
    static char 		 Name[BUFSIZ];
    MYdev_info_t		*Ptr;

    /*
     * If node name is a valid pointer, read the name from kernel space
     * and call probe routine to handle checking the device.
     */
    if (DevPtr->devi_name) {
	if (KVMread(kd, (u_long) DevPtr->devi_name, (char *) Name, 
		    sizeof(Name))) {
	    if (Debug) Error("Cannot read openprom device name.");
	    Name[0] = C_NULL;
	} else {
	    DevPtr->devi_name = (char *) strdup(Name);
	    OBPKcheckDevice(DevPtr, Parent, TreePtr);
	}
    }

    /*
     * If this node has children, read the child data from kernel space
     * and DESCEND.
     */
    if (DEVI_CHILD(DevPtr)) {
	Ptr = (MYdev_info_t *) xcalloc(1, sizeof(MYdev_info_t));
	if (KVMread(kd, (u_long) DEVI_CHILD(DevPtr), (char *) Ptr,
		    sizeof(MYdev_info_t))) {
	    Error("Cannot read openprom slave data for %s.", Name);
	} else {
	    DEVI_CHILD(DevPtr) = (MYdev_info_t *) Ptr;
	    OBPKtraverse(DEVI_CHILD(DevPtr), DevPtr, TreePtr);
	}
    }

    /*
     * If this node has a sibling, read the sibling data from kernel space
     * and TRAVERSE.
     */
    if (DEVI_SIBLING(DevPtr)) {
	Ptr = (MYdev_info_t *) xcalloc(1, sizeof(MYdev_info_t));
	if (KVMread(kd, (u_long) DEVI_SIBLING(DevPtr), (char *) Ptr,
		    sizeof(MYdev_info_t))) {
	    Error("Cannot read openprom next data for %s.", Name);
	} else {
	    DEVI_SIBLING(DevPtr) = (MYdev_info_t *) Ptr;
	    OBPKtraverse(DEVI_SIBLING(DevPtr), Parent, TreePtr);
	}
    }

    return(0);
}

/*
 * Get next OBP node id using "what" as what to get.
 */
static OBPnodeid_t OBPIOnext(NodeID, obp_fd, What)
    OBPnodeid_t		        NodeID;
    int				obp_fd;
    int				What;
{
    OBPio_t			opio;
    struct openpromio	       *op;
    int			       *idptr;

    op = &(opio.opio_oprom);
    idptr = (OBPnodeid_t *) (op->oprom_array);
    BZERO(opio.opio_buff, sizeof(opio.opio_buff));
    *idptr = NodeID;
    op->oprom_size = sizeof(opio.opio_buff);

    if (ioctl(obp_fd, What, op) < 0) {
	if (Debug) Error("OBP ioctl %s failed: %s.", 
			 (What == OPROMNEXT) ? "OPROMNEXT" : "OPROMCHILD", 
			 SYSERR);
	return(0);
    }

    return(*((OBPnodeid_t *)op->oprom_array));
}

/*
 * Determine if opvalue is a string or not.
 */
static int IsString(opvalue, opsize)
    char		       *opvalue;
    int				opsize;
{
    register int		i;

    for (i = 0; i < opsize - 1; ++i) {
	if (opvalue[i] == C_NULL)
	    return(0);
	if (iscntrl(opvalue[i]) || !isascii(opvalue[i]))
	    return(0);
    }

    return(1);
}

static char *GetLongValStr(String)
    char		       *String;
{
    static char			buff[BUFSIZ];
    long			lval;
    long		       *lptr;
    long			nval;

    lval = strtol(String, (char **)NULL, 0);
    lptr = (long *) &String[0];
    nval = *lptr;
    (void) sprintf(buff, "%u", (u_long) nval);

    return(buff);
}

/*
 * Decode an OBP value into a printable string.
 */
static char *OBPdecodeval(opvalue, opsize)
    char		       *opvalue;
    int				opsize;
{
    static char			buff[BUFSIZ];
    static char			buff2[2];
    register int		i;

    if (opsize == 0 || !opvalue)
	return((char *) NULL);

    if (IsString(opvalue, opsize)) {
	(void) strcpy(buff, opvalue);
	return(buff);
    } else {
	if (opsize > 4) {
	    (void) strcpy(buff, "0x");
	    for (i = 0; i < opsize; ++i) {
		if (i > 0 && ((i % 4) == 0))
		    (void) strcat(buff, ".0x");
		(void) sprintf(buff2, "%02x", opvalue[i] & 0xff);
		(void) strcat(buff, buff2);
	    }
	} else
	    (void) strcpy(buff, GetLongValStr(opvalue));
	return(buff);
    }
}

/*
 * Get the property value of propname from the OBP.
 */
static char *OBPgetpropval(propname, obp_fd)
    char		       *propname;
    int				obp_fd;
{
    static OBPio_t		opio;
    struct openpromio	       *op;

    op = &(opio.opio_oprom);
    op->oprom_size = sizeof(opio.opio_buff);
    (void) strcpy(op->oprom_array, propname);

    if (ioctl(obp_fd, OPROMGETPROP, op) < 0) {
	if (Debug) Error("OBP ioctl OPROMGETPROP failed for '%s': %s.", 
			 propname, SYSERR);
	return((char *)NULL);
    }

    if (op->oprom_size == -1) {
	if (Debug) Error("OBP no data available for '%s'.", propname);
	return((char *)NULL);
    }

    /*
     * op->oprom_array points at opio.opio_buff where the results
     * really live.
     */
    return(OBPdecodeval(op->oprom_array, op->oprom_size));
}

/*
 * Create and copy a OBPtable_t
 */
static OBPtable_t *CopyOBPtable(old)
    OBPtable_t	       *old;
{
    OBPtable_t	       *new;

    new = (OBPtable_t *) xmalloc(sizeof(OBPtable_t));
    BCOPY((char *)old, (char *)new, sizeof(OBPtable_t));

    return(new);
}

/*
 * Get all OBP properteries for the current OBP node and return
 * a table of the results.
 */
static OBPtable_t *OBPIOgetproptab(obp_fd)
    int				obp_fd;
{
    static OBPio_t		opio;
    struct openpromio	       *op;
    static OBPtable_t	       *OBPtable = NULL;
    static OBPtable_t		tab;
    register OBPtable_t        *tptr, *tabptr;
    register char	       *cp;
    register int		found;

    op = &(opio.opio_oprom);
    BZERO(opio.opio_buff, sizeof(opio.opio_buff));

    /*
     * Mark all previously allocated table entries as available
     */
    if (OBPtable)
	for (tptr = OBPtable; tptr; tptr = tptr->obpt_next)
	    tptr->obpt_key[0] = tptr->obpt_value[0] = C_NULL;

    for ( ; ; ) {
	/*
	 * Get next property
	 */
	op->oprom_size = sizeof(opio.opio_buff);
	if (ioctl(obp_fd, OPROMNXTPROP, op) < 0) {
	    if (Debug) Error("OBP ioctl OPROMNXTPROP failed: %s.", SYSERR);
	    return((OBPtable_t *) NULL);
	}

	/*
	 * We're done
	 */
	if (!op->oprom_size)
	    break;

	if (op->oprom_size >= sizeof(tab.obpt_key)) {
	    if (Debug) Error("Openprom key \"%s\" too large for buffer.",
			     op->oprom_array);
	    continue;
	}
	(void) strcpy(tab.obpt_key, op->oprom_array);
	tab.obpt_value[0] = C_NULL;

	/*
	 * Get the property value
	 */
	if (cp = OBPgetpropval(op->oprom_array, obp_fd)) {
	    if (strlen(cp) >= sizeof(tab.obpt_value)) {
		if (Debug) Error("Openprom value \"%s\" too large for buffer.",
				 cp);
		continue;
	    }
	    (void) strcpy(tab.obpt_value, cp);
	}

	/*
	 * Add this entry to the linked list.
	 * We try to use all previously allocated members of the linked
	 * list.  If none are available, we create a new member and add
	 * it to the list.  Allocated, but unused members have empty
	 * obpt_key and obpt_value strings.
	 */
	tab.obpt_next = NULL;
	if (OBPtable) {
	    for (tptr = OBPtable, found = FALSE; 
		 tptr && tptr->obpt_next && !found;
		 tptr = tptr->obpt_next) {

		/*
		 * See if this is an unused, but allocated entry
		 */
		if (tptr->obpt_key[0] == C_NULL &&
		    tptr->obpt_value[0] == C_NULL) {
		    tab.obpt_next = tptr->obpt_next;
		    BCOPY((char *) &tab, (char *) tptr, sizeof(tab));
		    found = TRUE;
		}
	    }
	    /*
	     * If we didn't find an unused entry slot, then create a new one
	     */
	    if (!found)
		tptr->obpt_next = CopyOBPtable(&tab);
	} else
	    OBPtable = CopyOBPtable(&tab);
    }

    return(OBPtable);
}

/*
 * Get OBP device data tab entry for "name"
 */
extern DEVDATATAB *OBPGetDevDataTab(Name)
    char 		       *Name;
{
    extern DEVDATATAB 	 	OBPdevices[];
    register int 		i;

    for (i = 0; OBPdevices[i].ddt_name; ++i)
	if (OBPdevices[i].ddt_flags & DDT_LENCMP) {
	    if (EQN(Name, OBPdevices[i].ddt_name, 
		    strlen(OBPdevices[i].ddt_name)))
		return(&OBPdevices[i]);
	} else {
	    if (EQ(Name, OBPdevices[i].ddt_name)) 
		return(&OBPdevices[i]);
	}

    return((DEVDATATAB *) NULL);
}

/*
 * Get the CPU model.
 *
 * The model is based upon the type of CPU and the clock frequency
 * of the CPU.
 */
static char *OBPgetCPUmodel(cputype, OBPtable, clockfreqptr)
    char		       *cputype;
    OBPtable_t		       *OBPtable;
    long		       *clockfreqptr;
{
    register OBPtable_t	       *tptr;
    long 			clockfreq = 0;
    register int		i;
    extern NAMETAB		CPUtypes[];

    /*
     * See if this node has it's own clock frequency
     */
    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	if (!tptr->obpt_key[0])
	    break;

	if (EQ(tptr->obpt_key, OBPK_CLOCKFREQ))
	    clockfreq = strtol(tptr->obpt_value, (char **)NULL, 0);
    }

    /*
     * No individual clock frequency, so we use the "global"
     * clock freqency that (hopefully) was set earlier.
     */
    if (!clockfreq)
	clockfreq = CpuClockFreq;

    /*
     * Set the frequency we're using so our caller can use it.
     */
    if (clockfreq && clockfreqptr)
	*clockfreqptr = clockfreq;

    /*
     * Scan the CPU types table looking for an entry that has
     * the same type name and the same clock frequency.
     */
    for (i = 0; CPUtypes[i].name; ++i)
	if (EQ(CPUtypes[i].name, cputype) && 
	    (clockfreq / MHERTZ) == CPUtypes[i].value)
	    return(CPUtypes[i].valuestr);

    /*
     * Try looking for just a type name match.
     */
    for (i = 0; CPUtypes[i].name; ++i)
	if (EQ(CPUtypes[i].name, cputype))
	    return(CPUtypes[i].valuestr);

    return((char *)NULL);
}

/*
 * Expand and OBP key string into something more easily read.
 */
static char *OBPexpandKey(string)
    char		       *string;
{
    register char	       *cp;
    register int		len;
    static char			buff[BUFSIZ];

    if (!string)
	return((char *)NULL);

    (void) strcpy(buff, string);

    /* Capitolize first letter */
    buff[0] = toupper(buff[0]);

    for (cp = &buff[0]; cp && *cp; ++cp) {
	if (*cp == '-' || *cp == '_') {
	    *cp++ = ' ';
	    if (*cp)
		*cp = toupper(*cp);
	}
    }

    /* Remove ending '?' */
    len = strlen(buff);
    if (buff[len-1] == '?')
	buff[len-1] = C_NULL;

    return(buff);
}

/*
 * Get a nice informative string describing OBPtab information.
 */
static char *OBPgetOBPinfo(OBPtab)
    OBPtable_t		       *OBPtab;
{
    register int		i;
    extern NAMETAB		OBPinfo[];
    static char			buff[BUFSIZ];
    char		       *FoundKey = NULL;

    /*
     * Try to see if this is one of the types of OBP info
     * that we want to use a more descriptive text than
     * the key itself.
     */
    for (i = 0; OBPinfo[i].name && !FoundKey; ++i)
	if (EQ(OBPinfo[i].name, OBPtab->obpt_key))
	    FoundKey = OBPinfo[i].valuestr;

    if (!FoundKey)
	FoundKey = OBPexpandKey(OBPtab->obpt_key);

    /* 
     * The ":" is used to force the device printing routing not to
     * interpret ":" as a key/value field seperator.
     */
    (void) strcpy(buff, ":");

    if (OBP_IS_BOOL(OBPtab))
	(void) strcat(buff, "Has ");
    (void) strcat(buff, FoundKey);
    if (!OBP_IS_BOOL(OBPtab)) {
	(void) strcat(buff, " is ");
	if (EQ(OBPtab->obpt_key, OBPK_CLOCKFREQ))
	    (void) strcat(buff, FreqStr(strtol(OBPtab->obpt_value,
					       (char **)NULL, 0)));
	else
	    (void) strcat(buff, OBPtab->obpt_value);
    }

    return(buff);
}

/*
 * Fix an OBP node name to be compatable with the
 * conventions used in SunOS 5.x.
 *
 * i.e. All '/' characters are mapped to '_'.
 */
static char *OBPfixNodeName(string)
    char		       *string;
{
    register char	       *cp;
    static char			buff[BUFSIZ];

    (void) strcpy(buff, string);
    cp = buff;
    while (cp = index(cp, '/'))
	*cp++ = '_';

    return(buff);
}

/*
 * Get (make) name of a CPU
 */
extern char *GetCpuName(ncpus, cpuid)
    int			       *ncpus;
    int				cpuid;
{
    static char			buff[BUFSIZ];

    (void) sprintf(buff, "cpu%d", (cpuid >= 0) ? cpuid : (*ncpus)++);

    return(buff);
}

/*
 * Probe a CPU using OBP info.
 */
extern DEVICE *OBPprobeCPU(NodeID, ParentID, OBPtable, TreePtr)
    OBPnodeid_t			NodeID;
    OBPnodeid_t			ParentID;
    OBPtable_t		       *OBPtable;
    DEVICE		      **TreePtr;
{
    register OBPtable_t	       *tptr;
    DEVICE		       *Device;
    DEVICE		       *fdev;
    register char	       *cp;
    register int		i;
    static char			buff[BUFSIZ];
    static int			listcnt;
    static int			ncpus = 0;
    int				cpuid = -1;
    char		       *cpumodel = NULL;
    long			clockfreq = 0;

    if (!(Device = NewDevice(NULL)))
	return((DEVICE *)NULL);

    listcnt = 0;
    Device->dv_desclist = (char **) xmalloc(MAXOBPDESCLIST * sizeof (char *));

    if (Debug) {
	(void) sprintf(buff, "Node ID is %d", NodeID);
	Device->dv_desclist[listcnt++] = strdup(buff);
	Device->dv_desclist[listcnt] = NULL;
    }

    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	if (!tptr->obpt_key[0])
	    break;

	/*
	 * See if this is the CPU type
	 */
	if (EQ(OBPK_NAME, tptr->obpt_key)) {
	    cpumodel = tptr->obpt_value;
	    Device->dv_model = OBPgetCPUmodel(cpumodel, OBPtable, &clockfreq);
	    continue;
	} else if (EQ(tptr->obpt_key, OBPK_CPUID))
	    cpuid = (int) strtol(tptr->obpt_value, (char **)NULL, 0);
	else if (EQ(tptr->obpt_key, OBPK_DEVTYPE) ||
		 EQ(tptr->obpt_key, OBPK_CLOCKFREQ))
	    continue;

	if (listcnt >= MAXOBPDESCLIST) {
	    if (Debug) Error("MAXOBPDESCLIST (%d) is too small.", 
			     MAXOBPDESCLIST);
	    continue;
	}
	Device->dv_desclist[listcnt++] = strdup(OBPgetOBPinfo(tptr));
	Device->dv_desclist[listcnt] = NULL;
    }

    if (cpumodel) {
	if (clockfreq)
	    (void) sprintf(buff, "%s %s", FreqStr(clockfreq), cpumodel);
	else
	    (void) strcpy(buff, cpumodel);
	Device->dv_desc = strdup(buff);
    }

    Device->dv_type = DT_CPU;
    Device->dv_nodeid = NodeID;
    if (cp = GetCpuName(&ncpus, cpuid))
	Device->dv_name = strdup(cp);

    /*
     * If no device by our name exists try to find a device
     * with a name which matches our CPU Model.  This should
     * be the root device node.  If so, make our device name
     * match the root device node name, so AddDevice() will 
     * add our info to the root device node later.
     */
    if (TreePtr && *TreePtr &&
	!FindDeviceByName(Device->dv_name, *TreePtr)) {

	fdev = FindDeviceByName(cpumodel, *TreePtr);
	if (!fdev) {
	    if (Debug) printf("Cannot find CPU device <%s>\n", cpumodel);
	    cp = OBPfixNodeName(cpumodel);
	    if (!EQ(cp, cpumodel))
		fdev = FindDeviceByName(cp, *TreePtr);
	} else
	    if (Debug) printf("Found CPU device <%s>\n", cpumodel);
	if (fdev && !fdev->dv_master)
	    Device->dv_name = fdev->dv_name;
    }

    return(Device);
}

/*
 * Get part description information using a model (part) name.
 */
extern PARTINFO *GetPartInfo(PartName)
    char		       *PartName;
{
    extern PARTINFO		PARTtable[];
    register int		i;
    register char	       *cp;

    for (i = 0; PARTtable[i].pt_part; ++i) {
	if (EQN(PARTtable[i].pt_part, PartName, strlen(PARTtable[i].pt_part)))
	    return(&PARTtable[i]);
	/* Try it without the manufacturer part of part name */
	if ((cp = index(PARTtable[i].pt_part, ',')) && 
	    EQN(++cp, PartName, strlen(cp)))
	    return(&PARTtable[i]);
    }

    return((PARTINFO *) NULL);
}

/*
 * This is for the SPARCcenter 2000 and SPARCserver 1000.
 *
 * If this device has a board number, then we make a "fake"
 * device called OBP_SYSBOARD that we insert in between the
 * Device and it's logical/current Master.
 */
static void OBPsetBoard(Device, BoardNum, TreePtr)
    DEVICE		       *Device;
    int				BoardNum;
    DEVICE		      **TreePtr;
{
    static char			BoardName[BUFSIZ];
    register DEVICE	       *DevPtr;
    register DEVICE	       *SaveDev;
    register DEVICE	       *LastDev;
    register DEVICE	       *OldMaster = NULL;
    register DEVICE	       *Master = NULL;
    DEVDATATAB		       *DevData = NULL;

    if (BoardNum < 0 || !Device)
	return;

    (void) sprintf(BoardName, "%s%d", OBP_SYSBOARD, BoardNum);

    /*
     * Find the Old Master
     */
    if (Device->dv_master) {
	OldMaster = FindDeviceByNodeID(Device->dv_master->dv_nodeid, 
				       *TreePtr);
	if (!OldMaster)
	    OldMaster = FindDeviceByName(Device->dv_master->dv_name, 
					 *TreePtr);
    }

    /*
     * If the Master (sysboard) device doesn't exit, create one.
     */
    if (!(Master = FindDeviceByName(BoardName, *TreePtr))) {
	Master = NewDevice(NULL);
	Master->dv_name = strdup(BoardName);
	Master->dv_unit = BoardNum;
	if (DevData = GetDevDataTab(OBP_SYSBOARD)) {
	    Master->dv_desc = DevData->ddt_desc;
	    Master->dv_model = DevData->ddt_model;
	    Master->dv_type = DevData->ddt_type;
	}
	if (OldMaster) {
	    Master->dv_mastername = OldMaster->dv_name;
	    Master->dv_master = OldMaster;
	}
	AddDevice(Master, TreePtr);
    }

    /*
     * Insert sysboard between Device and it's logical Master.
     */
    if (OldMaster) {
	Master->dv_master = OldMaster;

	/*
	 * Find Device in OldMaster and remove it.
	 */
	SaveDev = NULL;
	for (DevPtr = OldMaster->dv_slaves, LastDev = DevPtr; DevPtr; 
	     LastDev = DevPtr, DevPtr = DevPtr->dv_nxt) {

	    if (!EQ(DevPtr->dv_name, Device->dv_name))
		continue;

	    SaveDev = DevPtr;
	    LastDev->dv_nxt = DevPtr->dv_nxt;
	    break;
	}

	/*
	 * Add SaveDev back to new Master
	 */
	SaveDev->dv_nxt = NULL;
	if (Master->dv_slaves) {
	    for (DevPtr = Master->dv_slaves; DevPtr && DevPtr->dv_nxt;
		 DevPtr = DevPtr->dv_nxt);
	    DevPtr->dv_nxt = SaveDev;
	} else
	    Master->dv_slaves = SaveDev;
    }

    /*
     * Update Device
     */
    if (Master) {
	Device->dv_mastername = Master->dv_name;
	Device->dv_master = Master;
    }
}

/*
 * Decode memory string into number of MBYTES.
 */
static int OBPGetMemSize(MemStr)
    char		       *MemStr;
{
    register char	       *cp;
    char		       *MemBuff;
    u_long			Amt = 0;

    if (strchr(MemStr, '.')) {
	MemBuff = strdup(MemStr);
	for (cp = strtok(MemBuff, "."); cp; cp = strtok((char *)NULL, "."))
	    Amt += strtol(cp, (char **) NULL, 0);
	(void) free(MemBuff);
    } else
	Amt = strtol(MemStr, (char **) NULL, 0);

    return(Amt / MBYTES);
}

/*
 * Probe an unknown entity using OBP.
 */
extern DEVICE *OBPprobeUnknown(NodeID, ParentID, OBPtable, TreePtr)
    OBPnodeid_t			NodeID;
    OBPnodeid_t			ParentID;
    OBPtable_t		       *OBPtable;
    DEVICE		      **TreePtr;
{
    register OBPtable_t	       *tptr;
    DEVICE		       *Device = NULL;
    static char		      **desclist;
    register char	      **cpp;
    static char			buff[BUFSIZ];
    static int			listcnt;
    int				DevID = -1;
    int				BoardNum = -1;
    char		       *NodeName = NULL;
    char		       *Part = NULL;
    PARTINFO		       *PartInfo = NULL;
    int				MemSize = 0;

    /*
     * First see if we can find this device by it's OBP node ID
     */
    if (NodeID && TreePtr)
	Device = FindDeviceByNodeID(NodeID, *TreePtr);

    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	if (!tptr->obpt_key[0])
	    break;

	if (EQ(OBPK_NAME, tptr->obpt_key))
	    NodeName = tptr->obpt_value;
    }

    /*
     * Try to find the device by name.   This usually only
     * works for the root "system" node and a few others.
     */
    if (!Device && NodeName && TreePtr)
	Device = FindDeviceByName(NodeName, *TreePtr);

    /*
     * Didn't find it.
     */
    if (!Device)
	return((DEVICE *)NULL);

    /*
     * Add any existing description info to the beginning of our
     * new list.
     */
    listcnt = 0;
    desclist = (char **) xmalloc(MAXOBPDESCLIST * sizeof (char *));
    if (Device)
	for (cpp = Device->dv_desclist; cpp && *cpp; ++cpp) {
	    desclist[listcnt++] = *cpp;
	    desclist[listcnt] = NULL;
	}

    /*
     * Get descriptive and other info
     */
    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	if (!tptr->obpt_key[0])
	    break;

	/*
	 * Skip these
	 */
	if (EQ(tptr->obpt_key, OBPK_NAME))
	    continue;
	else if (EQ(tptr->obpt_key, OBPK_MODEL))
	    Part = tptr->obpt_value;
	else if (EQ(tptr->obpt_key, OBPK_BOARD))
	    BoardNum = strtol(tptr->obpt_value, (char **) NULL, 0);
	else if (EQ(tptr->obpt_key, OBPK_DEVICEID))
	    DevID = strtol(tptr->obpt_value, (char **) NULL, 0);
	else if (EQ(tptr->obpt_key, OBPK_SIZE) && 
		 EQ(NodeName, OBPK_MEMUNIT))
	    MemSize = OBPGetMemSize(tptr->obpt_value);

	if (listcnt >= MAXOBPDESCLIST) {
	    if (Debug) Error("MAXOBPDESCLIST (%d) is too small.", 
			     MAXOBPDESCLIST);
	    continue;
	}
	desclist[listcnt++] = strdup(OBPgetOBPinfo(tptr));
	desclist[listcnt] = NULL;
    }

    /*
     * Handle special cases
     */
    if (MemSize && Device->dv_model) {
	(void) sprintf(buff, "%d MB %s", MemSize, Device->dv_model);
	Device->dv_model = strdup(buff);
    }
    if (BoardNum >= 0)
	OBPsetBoard(Device, BoardNum, TreePtr);

    /*
     * Add information to existing Device.
     */

    if (Part) {
	PartInfo = GetPartInfo(Part);
	if (PartInfo) {
	    Device->dv_model = PartInfo->pt_name;
	    Device->dv_desc = PartInfo->pt_desc;
	} else if (Device->dv_model) {
	    (void) sprintf(buff, "%s (%s)", Device->dv_model, Part);
	    Device->dv_model = strdup(buff);
	} else
	    Device->dv_model = strdup(Part);
    }

    Device->dv_desclist = desclist;

    return((DEVICE *)NULL);
}

/*
 * Check over an OBP node.
 */
static char *OBPIOcheckNode(NodeID, obp_fd, ParentName, ParentID, TreePtr)
    OBPnodeid_t			NodeID;
    int				obp_fd;
    char		       *ParentName;
    OBPnodeid_t			ParentID;
    DEVICE		      **TreePtr;
{
    OBPtable_t		       *OBPtable;
    register OBPtable_t	       *tptr;
    static int			first = TRUE;
    static char			MyName[BUFSIZ];
    static char			buff[BUFSIZ];
    DEVDATATAB		       *devtab = NULL;
    DEVICE		       *device = NULL;
    DEVICE		       *master = NULL;

    /*
     * Get all the properties for the current node.
     */
    if (!(OBPtable = OBPIOgetproptab(obp_fd)))
	return((char *)NULL);

    /*
     * Iterate over all the properties looking for interesting properties
     * or until we find the device type.  If the device type is one which
     * we found in our probe list, go ahead and probe it.
     */
    MyName[0] = C_NULL;
    for (tptr = OBPtable; tptr; tptr = tptr->obpt_next) {
	/*
	 * Have we reached the end of the valid list?
	 */
	if (!tptr->obpt_key[0])
	    break;

	if (Debug) printf("OBPcheckNode: NodeID %d <%s> = <%s> (0x%x)\n",
			  NodeID,
			  tptr->obpt_key, tptr->obpt_value,
			  atol(tptr->obpt_value));

	/*
	 * If this is the first node and this property is the Clock Frequency
	 * then save it for later use by the individual CPUs.
	 */
	if (first && !CpuClockFreq && EQ(tptr->obpt_key, OBPK_CLOCKFREQ))
	    CpuClockFreq = strtol(tptr->obpt_value, (char **)NULL, 0);

	if (EQ(tptr->obpt_key, OBPK_DEVTYPE))
	    devtab = OBPGetDevDataTab(tptr->obpt_value);

	if (EQ(tptr->obpt_key, OBPK_NAME) && tptr->obpt_value[0])
	    (void) strcpy(MyName, tptr->obpt_value);
    }

    if (devtab)
	device = (*devtab->ddt_probe)(NodeID, ParentID, OBPtable, TreePtr);

    if (!device)
	device = OBPprobeUnknown(NodeID, ParentID, OBPtable, TreePtr);

    if (device) {
	/*
	 * Find or create of Master
	 */
	if (!device->dv_master) {
	    master = FindDeviceByNodeID(ParentID, *TreePtr);
	    if (!master && ParentName) {
		master = NewDevice(NULL);
		master->dv_name = strdup(ParentName);
		master->dv_nodeid = ParentID;
	    }
	    device->dv_master = master;
	}
	device->dv_nodeid = NodeID;
	AddDevice(device, TreePtr);
    }

    first = FALSE;

    return((device) ? device->dv_name : ((MyName[0]) ? MyName : (char *)NULL));
}

/*
 * Traverse OBP I/O nodes
 */
static int OBPIOtraverse(NodeID, obp_fd, ParentName, ParentID, TreePtr)
    OBPnodeid_t		        NodeID;
    int				obp_fd;
    char		       *ParentName;
    OBPnodeid_t		        ParentID;
    DEVICE		      **TreePtr;
{
    OBPnodeid_t		        NewID;
    char		       *MyName;
    char		       *cp;

    if (cp = OBPIOcheckNode(NodeID, obp_fd, ParentName, ParentID, TreePtr))
	MyName = strdup(cp);

    if (NewID = OBPIOnext(NodeID, obp_fd, OPROMCHILD))
	OBPIOtraverse(NewID, obp_fd, MyName, NodeID, TreePtr);

    if (NewID = OBPIOnext(NodeID, obp_fd, OPROMNEXT))
	OBPIOtraverse(NewID, obp_fd, ParentName, ParentID, TreePtr);

    return(0);
}

/*
 * Build devices based on what we find using raw I/O calls
 * to the OBP.
 */
static int OBPIObuild(TreePtr)
    DEVICE		      **TreePtr;
{
    int				obp_fd;
    int				status;

    /*
     * Note: Only one process at a time can open _PATH_OPENPROM.
     */
    if ((obp_fd = open(_PATH_OPENPROM, O_RDONLY)) < 0) {
	if (Debug) Error("Cannot open \"%s\": %s.", _PATH_OPENPROM, SYSERR);
	return(-1);
    }

    status = OBPIOtraverse(OBPIOnext((OBPnodeid_t) 0, obp_fd, OPROMNEXT),
			   obp_fd, (char *)NULL, 0, TreePtr);
    (void) close(obp_fd);

    return(status);
}

/*
 * Find and build devices based on what we find from the OBP
 * both from the kernel and the OPENPROM itself.
 */
extern int BuildOpenPROM(TreePtr)
    DEVICE		      **TreePtr;
{
    int				found = 1;

    if (OBPKbuild(TreePtr) == 0)
	found = 0;

    if (OBPIObuild(TreePtr) == 0)
	found = 0;

    return(found);
}

/*
 * Get the CPU type by reading and returning the name of the
 * root (top) OBP node from the kernel.
 */
static char *OBPKgetCpuType()
{
    MYdev_info_t	       *Root;
    static char			Name[BUFSIZ];

    if (!(Root = OBPKgetRoot()))
	return((char *) NULL);
    
    if (!Root->devi_name) {
	if (Debug) Error("OBP Root node does not have a name.");
	return((char *) NULL);
    }

    /*
     * Read the name from kernel space.
     */
    if (KVMread(kd, (u_long) Root->devi_name, (char *) Name, sizeof(Name))) {
	if (Debug) Error("Cannot read OBP root device name.");
	return((char *) NULL);
    }

    if (Name && *Name)
	return(Name);
    else
	return((char *) NULL);
}

/*
 * Get system model type from OBP directly.
 */
static char *OBPIOgetCpuType()
{
    int				obp_fd;
    char		       *Name;

    /*
     * Note: Only one process at a time can open _PATH_OPENPROM.
     */
    if ((obp_fd = open(_PATH_OPENPROM, O_RDONLY)) < 0) {
	if (Debug) Error("Cannot open \"%s\": %s.", _PATH_OPENPROM, SYSERR);
	return((char *) NULL);
    }

    Name = OBPIOcheckNode(OBPIOnext((OBPnodeid_t) 0, obp_fd, OPROMNEXT),
			  obp_fd, (char *) NULL, (OBPnodeid_t) 0, 
			  (DEVICE *) NULL);

    (void) close(obp_fd);

    return(Name);
}

/*
 * Get system CPU type from OBP.
 */
extern char *OBPGetCpuType()
{
    char		       *Name;

    /*
     * First try getting it from the kernel, if that
     * fails, then query the OBP directly.
     */
    Name = OBPKgetCpuType();
    if (!Name)
	Name = OBPIOgetCpuType();

    if (Name)
	return(OBPcleanName(Name, TRUE));
    else
	return((char *) NULL);
}

#endif	/* HAVE_OPENPROM */
