/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/os-sunos.c,v 1.50 1993/06/08 18:15:07 mcooper Exp mcooper $";
#endif

/*
 * General SunOS specific routines
 */

#include "defs.h"

#include <nlist.h>
#include <sys/stat.h>

/*
 * Name of frame buffer "indirect" device.
 */
#define FBDEVICE		"fb"

#if	SUNOS == 5
#include "os-sunos5.h"
typedef short			cputype_t;
#else
#include "os-sunos4.h"
typedef int			cputype_t;
#endif	/* SUNOS == 5 */

struct stat			StatBuf;
cputype_t			CpuType = 0;
static DEVICE 		       *Device;
static char 			Buf[BUFSIZ];
kvm_t		       	       *kd;
extern char		        CpuSYM[];
extern char		        IdpromSYM[];

/*
 * Build device tree using TreePtr.
 * Calls bus and method specific functions to
 * search for devices.
 */
extern int BuildDevicesSunOS(TreePtr)
    DEVICE 		       **TreePtr;
{
    int 			 Found = 1;

#if	defined(HAVE_OPENPROM)
    if (BuildOpenPROM(TreePtr) == 0)
	Found = 0;
#endif	/* HAVE_OPENPROM */

#if	defined(HAVE_MAINBUS)
    if (BuildMainBus(TreePtr) == 0)
	Found = 0;
#endif	/* HAVE_MAINBUS */

    return(Found);
}

/*
 * Scan the Disk Controller table looking for
 * a specific type.
 */
extern DKCTLRTAB *GetDkCtlrTab(DkCtrlType)
    int 			DkCtrlType;
{
    extern DKCTLRTAB 		DkCtlrTab[];
    register int 		i;

    for (i = 0; DkCtlrTab[i].ct_model; ++i) {
	if (DkCtrlType == DkCtlrTab[i].ct_ctype)
	    return(&DkCtlrTab[i]);
    }

    return((DKCTLRTAB *) NULL);
}

/*
 * Scan the Frame Buffer table looking for 
 * a specific fb type.
 */
extern NAMETAB *GetFBTab(FBType)
    int 			FBType;
{
    extern NAMETAB 		FBTab[];
    register int 		i;

    for (i = 0; FBTab[i].name; ++i) {
	if (FBType == FBTab[i].value)
	    return(&FBTab[i]);
    }

    return((NAMETAB *) NULL);
}

/*
 * Get the cpu type from the kernel
 */
cputype_t GetCpuType()
{
    struct nlist	       *nlptr;
    kvm_t		       *kd;
    cputype_t			cpu;

    if (!(kd = KVMopen()))
	return(-1);

    if ((nlptr = KVMnlist(kd, CpuSYM, (struct nlist *)NULL)) == NULL)
	return(-1);

    if (CheckNlist(nlptr))
	return(-1);

    if (KVMread(kd, (u_long) nlptr->n_value, (char *) &cpu, sizeof(cpu))) {
	if (Debug) Error("Cannot read cpu type from kernel.");
	return(-1);
    }

    KVMclose(kd);

    return(cpu);
}

/*
 * Determine our cpu model name.
 *
 * We first try to find the "cpu" or "cputype" symbol in the kernel.
 * If that's not present or is of an unknown value, we use the OBP
 * root node name.
 */
extern char *GetModelName()
{
    extern NAMETAB 		ModelTab[];
    register int 		i;
    register char	       *cp = NULL;
    register char	       *Name = NULL;

    if (CpuType <= 0 && !UseProm)
	CpuType = GetCpuType();

    if (CpuType > 0 && !UseProm)
	for (i = 0; ModelTab[i].name; ++i)
	    if (CpuType == ModelTab[i].value) {
		Name = ModelTab[i].name;
		break;
	    }

#if	defined(HAVE_OPENPROM)
    /*
     * SunOS no longer identifies individual machines by the "cputype"
     * symbol in the kernel.  Instead, the root OBP node name is used.
     * This started with the SPARCclassic, LX, and SPARCcenter 2000.
     */
    if (!Name || UseProm)
	if (cp = OBPGetCpuType())
	    Name = strdup(cp);
#endif	/* HAVE_OPENPROM */

    if (Debug) printf("CPU = 0x%x  Name = <%s>\n", CpuType, ARG(Name));

    return(Name);
}

#if 	defined(CPU_ARCH) /* Sun */
#define	ARCH_MASK CPU_ARCH
#endif	/* CPU_ARCH */
#if 	defined(CPU_TYPE) /* Solbourne */
#define ARCH_MASK CPU_TYPE
#endif	/* CPU_TYPE */
/*
 * Determine our kernel architecture name from our hostid.
 */
extern char *GetKernArchName()
{
#if	defined(ARCH_MASK)
    extern NAMETAB 		KernArchTab[];
    register int 		i;

    if (!CpuType)
	if ((CpuType = GetCpuType()) < 0)
	    return((char *) NULL);

    for (i = 0; KernArchTab[i].name; ++i)
	if ((CpuType & ARCH_MASK) == KernArchTab[i].value)
	    return(KernArchTab[i].name);

    if (Debug)
	Error("Kernel Arch 0x%x not defined; Cpu = 0x%x Mask = 0x%x", 
	      CpuType & ARCH_MASK, CpuType, ARCH_MASK);
#endif	/* ARCH_MASK */

    return((char *) NULL);
}

#if	defined(HAVE_IDPROM)
/*
 * Get system serial number
 */
static struct idprom *GetIDPROM()
{
    static struct idprom	idprom;
    struct nlist	       *nlptr;
    kvm_t		       *kd;

    if (!(kd = KVMopen()))
	return((struct idprom *)NULL);

    if ((nlptr = KVMnlist(kd, IdpromSYM, (struct nlist *)NULL)) == NULL)
	return((struct idprom *)NULL);

    if (CheckNlist(nlptr))
	return((struct idprom *)NULL);

    if (KVMread(kd, (u_long) nlptr->n_value, (char *) &idprom, 
		sizeof(idprom))) {
	if (Debug) Error("Cannot read \"%s\" from kernel.", IdpromSYM);
	return((struct idprom *)NULL);
    }

    KVMclose(kd);

    return(&idprom);
}
#endif	/* HAVE_IDPROM */

/*
 * Get system serial number
 */
extern char *GetSerialNoStr()
{
    static char			buff[BUFSIZ];
#if	defined(HAVE_IDPROM)
    struct idprom	       *idprom;

    if (!(idprom = GetIDPROM()))
	return((char *)NULL);

    /*
     * id_format is not set correctly under SunOS 4.x
     */
    if (idprom->id_format != IDFORM_1)
	if (Debug) Error("Warning: IDPROM format (%d) is incorrect.",
			 idprom->id_format);

    (void) sprintf(buff, "%d", idprom->id_serial);

#endif	/* HAVE_IDPROM */
    return((buff[0]) ? buff : (char *)NULL);
}

/*
 * Get version of OS
 */
extern char *GetOSVersionStr()
{
    return((char *)NULL);
}

/*
 * Probe a FrameBuffer.
 */
extern DEVICE *ProbeFrameBuffer(fbname, DevData, DevDataTab)
    char 		       *fbname;
    DEVDATA 		       *DevData;
    DEVDATATAB	 	       *DevDataTab;
{
#if	defined(FBIOMONINFO)
    struct mon_info		mon_info;
#endif	/* FBIOMONINFO */
#if	defined(FBIOGXINFO)
    struct cg6_info 		cg6_info;
#endif	/* FBIOGXINFO */
    struct fbgattr		fbgattr;
    DEVICE 		       *fbdevice;
    FRAMEBUFFER 	       *fb;
    NAMETAB 		       *fbtab;
    char 		       *file;
    static char			filebuff[MAXPATHLEN];
    static char			buff[BUFSIZ], buff2[BUFSIZ];
    char		      **desclist = NULL;
    int				listcnt = 0;
    int 			desc;
    register int		i;

    if (!fbname)
	return((DEVICE *) NULL);

    if (Debug)
	printf("ProbeFrameBuffer '%s'\n", fbname);

#if	SUNOS == 5
    if (EQ(fbname, FBDEVICE))
	(void) sprintf(filebuff, "%s/%s", _PATH_DEV, fbname);
    else
	(void) sprintf(filebuff, "%s/%s", _PATH_DEV_FBS, fbname);
    file = filebuff;
#else
    file = GetCharFile(fbname, NULL);
#endif	/* SUNOS == 5 */

    /*
     * Check the device file.  If the stat fails because
     * the device doesn't exist, trying the default framebuffer
     * device /dev/fb.
     */
    if (stat(file, &StatBuf) != 0) {
	if (errno == ENOENT && !EQ(fbname, FBDEVICE)) {
	    if (Debug) 
		Error("Framebuffer device `%s' does not exist.  Trying `%s'.",
		      fbname, FBDEVICE);
	    return(ProbeFrameBuffer(FBDEVICE, DevData, DevDataTab));
	}
    }

    if ((desc = open(file, O_RDONLY)) < 0) {
	if (Debug) Error("%s: Cannot open for reading: %s.", file, SYSERR);
	return((DEVICE *) NULL);
    }

    /*
     * Get real fb attributes
     */
    if (ioctl(desc, FBIOGATTR, &fbgattr) != 0) {
	if (Debug) Error("%s: FBIOGATTR failed: %s.", file, SYSERR);
	if (ioctl(desc, FBIOGTYPE, &fbgattr.fbtype) != 0) {
	    if (Debug) Error("%s: FBIOGTYPE failed: %s.", file, SYSERR);
	    return((DEVICE *) NULL);
	}
    }

    /*
     * We're committed to try
     */
    if (!(fb = NewFrameBuffer(NULL))) {
	Error("Cannot create new frame buffer.");
	return((DEVICE *) NULL);
    }

    if (!(fbdevice = NewDevice(NULL))) {
	Error("Cannot create new frame buffer device entry.");
	return((DEVICE *) NULL);
    }

    /*
     * Allocate one more than we need
     */
    desclist 			= (char **)xcalloc(8, sizeof(char *));

#if	defined(FBIOGXINFO)
    /*
     * GX (cgsix) info
     */
    if (ioctl(desc, FBIOGXINFO, &cg6_info) == 0) {
	(void) sprintf(buff, "SBus Slot %d, Revision %d",
		       cg6_info.slot, cg6_info.boardrev);
	if (cg6_info.hdb_capable)
	    (void) strcat(buff, ", double buffered");
	else
	    (void) strcat(buff, ", single buffered");
	desclist[listcnt++] = strdup(buff);
	if (cg6_info.vmsize)
	    fb->fb_vmsize 	= mbytes_to_bytes(cg6_info.vmsize);
    } else
	if (Debug) Error("%s: FBIOGXINFO failed: %s.", file, SYSERR);
#endif 	/* FBIOGXINFO */

#if	defined(FBIOMONINFO)
    /*
     * Monitor info
     */
    if (ioctl(desc, FBIOMONINFO, &mon_info) == 0) {
	(void) sprintf(buff, "Monitor: Pixel Frequency is %s",
		       FreqStr(mon_info.pixfreq));
	desclist[listcnt++] = strdup(buff);

	(void) sprintf(buff, "Monitor: Horizontal Frequency is %s",
		       FreqStr(mon_info.hfreq));
	desclist[listcnt++] = strdup(buff);

	(void) sprintf(buff, "Monitor: Vertical Frequency is %s",
		       FreqStr(mon_info.vfreq));
	desclist[listcnt++] = strdup(buff);

	(void) sprintf(buff, "Monitor: Horizontal Sync is %d pixels",
		       mon_info.hsync);
	desclist[listcnt++] = strdup(buff);

	(void) sprintf(buff, "Monitor: Vertical Sync is %d scanlines",
		       mon_info.vsync);
	desclist[listcnt++] = strdup(buff);
    } else
	if (Debug) Error("%s: FBIOMONINFO failed: %s.", file, SYSERR);
#endif	/* FBIOMONINFO */

    /*
     * We're done doing ioctl()'s
     */
    close(desc);

    /*
     * Find out what type of fb this is.
     */
    if (fbtab = GetFBTab(fbgattr.fbtype.fb_type)) {
	if (fbtab->valuestr) {
	    (void) sprintf(buff, "%s (%s)", fbtab->name, fbtab->valuestr); 
	    fbdevice->dv_model		= strdup(buff);
	} else
	    fbdevice->dv_model 		= fbtab->name;
    } else {
	Error("Device `%s' is an unknown type (%d) of frame buffer.",
	      fbname, fbgattr.fbtype.fb_type);
	fbdevice->dv_model 		= "UNKNOWN";
    }

#if	defined(FB_ATTR_NEMUTYPES)
    /*
     * See if this fb emulates other fb's.
     */
    buff[0] = C_NULL;
    for (i = 0; i < FB_ATTR_NEMUTYPES && fbgattr.emu_types[i] >= 0; ++i)
	if (fbgattr.emu_types[i] != fbgattr.fbtype.fb_type &&
	    (fbtab = GetFBTab(fbgattr.emu_types[i]))) {
	    if (buff[0])
		(void) strcat(buff, ", ");
	    (void) strcat(buff,
			  (fbtab->valuestr) ? fbtab->valuestr : fbtab->name);
	}
    if (buff[0]) {
	(void) sprintf(buff2, "Emulates %s", buff);
	desclist[listcnt++]		= strdup(buff2);
    }
#endif	/* FB_ATTR_NEMUTYPES */

    /*
     * Put things together
     */
    fbdevice->dv_name 			= fbname;
    fbdevice->dv_type 			= DT_FRAMEBUFFER;
    fbdevice->dv_nodeid			= DevData->dd_nodeid;
    fbdevice->dv_devspec 		= (caddr_t *) fb;
    if (desclist && desclist[0]) {
	desclist[listcnt]		= NULL;
	fbdevice->dv_desclist		= desclist;
    } else if (desclist)
	(void) free(desclist);

    fb->fb_height 			= fbgattr.fbtype.fb_height;
    fb->fb_width 			= fbgattr.fbtype.fb_width;
    fb->fb_depth 			= fbgattr.fbtype.fb_depth;
    fb->fb_size 			= fbgattr.fbtype.fb_size;
    fb->fb_cmsize 			= fbgattr.fbtype.fb_cmsize;

    fbdevice->dv_master 		= MkMasterFromDevData(DevData);

    return(fbdevice);
}
