/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/os-sunos5.c,v 1.11 1993/06/08 18:18:29 mcooper Exp mcooper $";
#endif

/*
 * SunOS 5.x specific routines
 */

#include "defs.h"
#include <nlist.h>
#include <dirent.h>
#include <sys/stat.h>

/*
 * CPU (model) symbol
 */
char 				CpuSYM[] = "_cputype";
static char			Buff[BUFSIZ];
DEVICE			       *Device;
extern char		        RomVecSYM[];
struct stat			StatBuf;

/*
 * No such device for SunOS 5 right now
 */
extern DEVICE *ProbeSPDrive()
{
    return((DEVICE *)NULL);
}

/*
 * This routine is obsolete.
 */
extern DEVICE *ProbeCDROMDrive()
{
    return((DEVICE *)NULL);
}

/* 
 * Scan DevDir looking for a device file that matches 
 * DevData->dd_devnum.
 */
static char *GetDeviceFile(DevData, DevDir)
    DEVDATA 		       *DevData;
    char		       *DevDir;
{
    static DIR		       *dirp = NULL;
    struct dirent	       *dirent;
    static char			path[MAXPATHLEN];
    static char			lastdir[MAXPATHLEN];

    /* Don't bother trying if the devnum is 0 */
    if (DevData->dd_devnum <= 0)
	return((char *)NULL);

    if (!dirp || !EQ(DevDir, lastdir)) {
	if (!(dirp = opendir(DevDir))) {
	    if (Debug) Error("Cannot open directory \"%s\": %s", 
			     DevDir, SYSERR);
	    return((char *)NULL);
	}
	(void) strcpy(lastdir, DevDir);
    } else
	rewinddir(dirp);

    while (dirent = readdir(dirp)) {
	if (EQ(dirent->d_name, ".") || EQ(dirent->d_name, ".."))
	    continue;
	
	(void) sprintf(path, "%s/%s", DevDir, dirent->d_name);

	if (stat(path, &StatBuf) < 0) {
	    if (Debug) Error("Stat of %s failed: %s.", path, SYSERR);
	    continue;
	}

#if	defined(DEBUG_VERBOSE)
	if (Debug) printf("GetDeviceFile dev '%s' (%d) got '%s' (%d)\n",
			  DevData->dd_devname, DevData->dd_devnum,
			  path, StatBuf.st_rdev);
#endif	/* DEBUG_VERBOSE */

	if (StatBuf.st_rdev == DevData->dd_devnum)
	    return(path);
    }

    return((char *)NULL);
}

/*
 * Probe a Tape Drive.
 */
extern DEVICE *ProbeTapeDrive(tapename, DevData, DevDataTab)
    /*ARGSUSED*/
    char 		       *tapename;
    DEVDATA 		       *DevData;
    DEVDATATAB	 	       *DevDataTab;
{
    extern NAMETAB		MtInfo[];
    DEVICE 		       *Device;
    char 		       *devfile;
    char 		       *devname;
    char 		       *model = NULL;
    static char 		devfilen[BUFSIZ];
    struct mtget 		mtget;
    register int 		i;
    register char	       *cp;
    int 			fd;

    if (!(devfile = GetDeviceFile(DevData, _PATH_DEV_RMT))) {
	if (Debug) Error("Cannot find tape device for `%s'", tapename);
	return((DEVICE *)NULL);
    }

    /* 
     * Use devfile to get name of device.
     * Takes something like "/dev/rmt/0l" and
     * turns it into "rmt/0".
     */
    cp = devfile;
    /* Skip over "/dev/" */
    if (strlen(devfile) > 5 && EQN(devfile, "/dev/", 4))
	cp += 5;
    devname = cp;
    if (cp = strrchr(devfile, '/')) {
	++cp;
	while (isdigit(*++cp));
	if (*cp && !isdigit(*cp))
	    *cp = C_NULL;
    }

    /* Use the no rewind file */
    (void) sprintf(devfilen, "%sn", devfile);

    if ((fd = open(devfilen, O_RDONLY)) < 0) {
	if (Debug)
	    Error("%s Cannot open for read: %s.", devfilen, SYSERR);

	/*
	 * If we know for sure this drive is present and we
	 * know something about it, then create a minimal device.
	 */
	if ((DevDataTab->ddt_model || DevDataTab->ddt_desc) &&
	    FLAGS_ON(DevData->dd_flags, DD_IS_ALIVE)) {
	    Device 			= NewDevice((DEVICE *) NULL);
	    Device->dv_name 		= strdup(devname);
	    Device->dv_aname 		= strdup(tapename);
	    Device->dv_unit 		= DevData->dd_devunit;
	    Device->dv_master 		= MkMasterFromDevData(DevData);
	    Device->dv_type 		= DT_TAPEDRIVE;
	    Device->dv_nodeid		= DevData->dd_nodeid;
	    Device->dv_model 		= DevDataTab->ddt_model;
	    Device->dv_desc 		= DevDataTab->ddt_desc;
	    return(Device);
	} else
	    return((DEVICE *) NULL);
    }

    if (ioctl(fd, MTIOCGET, &mtget) != 0) {
	Error("%s: Cannot extract tape status: %s.", devname, SYSERR);
	return((DEVICE *) NULL);
    }

    (void) close(fd);

    model = "unknown";
    for (i = 0; MtInfo[i].name; ++i) {
	if ((MtInfo[i].value == mtget.mt_type)) {
	    model = MtInfo[i].name;
	    break;
	}
    }

    /*
     * Create and set device info
     */
    Device 			= NewDevice(NULL);
    Device->dv_name 		= strdup(devname);
    Device->dv_aname 		= strdup(tapename);
    Device->dv_type 		= DT_TAPEDRIVE;
    Device->dv_nodeid		= DevData->dd_nodeid;
    if (model)
	Device->dv_model 	= model;
    else
	Device->dv_model 	= DevDataTab->ddt_model;
    Device->dv_desc 		= DevDataTab->ddt_desc;
    Device->dv_unit 		= DevData->dd_devunit;
    Device->dv_master 		= MkMasterFromDevData(DevData);

    return(Device);
}

/*
 * Get SVR4 style partition information
 */
DKvtoc *GETvtoc(fd, devfile)
    int				fd;
    char		       *devfile;
{
    static DKvtoc 		vtoc;

    if (ioctl(fd, DKIOCGVTOC, &vtoc) < 0) {
	if (Debug) Error("%s: DKIOCGVTOC: %s.", devfile, SYSERR);
	return((DKvtoc *)NULL);
    }

    return(&vtoc);
}

/*
 * Get controller information
 */
DKcinfo *GETdk_cinfo(fd, devfile)
    int				fd;
    char		       *devfile;
{
    static DKcinfo 		dk_cinfo;

    if (ioctl(fd, DKIOCINFO, &dk_cinfo) < 0) {
	if (Debug) Error("%s: DKIOCINFO: %s.", devfile, SYSERR);
	return((DKcinfo *)NULL);
    }

    return(&dk_cinfo);
}

/*
 * Get disk geometry
 */
DKgeom *GETdk_geom(fd, devfile)
    int				fd;
    char		       *devfile;
{
    static DKgeom 		dk_geom;

    if (ioctl(fd, DKIOCGGEOM, &dk_geom) < 0) {
	if (Debug) Error("%s: DKIOCGGEOM: %s.", devfile, SYSERR);
	return((DKgeom *)NULL);
    }

    return(&dk_geom);
}

/*
 * Get disk type structure.
 */
static DKtype *GETdk_type(fd, file)
    int 			fd;
    char 		       *file;
{
    static DKtype 		dk_type;

    if (ioctl(fd, HDKIOCGTYPE, &dk_type) < 0) {
	if (errno != ENOTTY)
	    if (Debug) Error("%s: DKIOCGTYPE: %s.", file, SYSERR);
	return(NULL);
    }

    return(&dk_type);
}

/*
 * Find mount point for a given device
 */
static char *GetMountInfo(devname, part)
    char		       *devname;
    char		       *part;
{
    static FILE 	       *mf = NULL;
    struct mnttab 	        mnttab;
    static char			name[BUFSIZ];
    register char	       *cp;

    if (!devname || !part)
	return((char *) NULL);

    (void) sprintf(name, "%s%s", devname, part);

    if (!mf) {
	if ((mf = fopen(MNTTAB, "r")) == NULL) {
	    Error("%s: Cannot open for reading: %s.", MNTTAB, SYSERR);
	    return(NULL);
	}
    } else
	rewind(mf);

    while (getmntent(mf, &mnttab) == 0) {
	if (!mnttab.mnt_special)
	    continue;
	if (cp = strrchr(mnttab.mnt_special, '/'))
	    ++cp;
	else
	    cp = mnttab.mnt_special;
	if (EQ(cp, name))
	    return(strdup(mnttab.mnt_mountp));
    }

    return((char *) NULL);
}

/*
 * Translate disk partition information from basic
 * extracted disk info.
 */
static DISKPART *GetDiskPart(devname, dk_vtoc)
    char 		       *devname;
    DKvtoc 		       *dk_vtoc;
{
    register DISKPART 	       *pdp;
    register ushort 		i;
    static char 		pname[3];
    DISKPART 		       *base = NULL;
    DISKPART			diskpart;

    if (!devname || !dk_vtoc)
	return((DISKPART *) NULL);

    pname[2] = C_NULL;
    /*
     * dk_vtoc->v_nparts is always 0 after the first time through
     * here, so we use V_NUMPAR.
     */
    for (i = 0; i < V_NUMPAR; ++i) {
	/* Skip partitions that have no size */
	if (!dk_vtoc->v_part[i].p_size)
	    continue;
	(void) sprintf(pname, "s%d", i);
	diskpart.dp_name = strdup(pname);
	diskpart.dp_mnt = GetMountInfo(devname, pname);
	diskpart.dp_stsect = dk_vtoc->v_part[i].p_start;
	diskpart.dp_nsect = dk_vtoc->v_part[i].p_size;

	if (base) {
	    for (pdp = base; pdp && pdp->dp_nxt; pdp = pdp->dp_nxt);
	    pdp->dp_nxt = NewDiskPart(&diskpart);
	} else
	    base = NewDiskPart(&diskpart);
    }

    return(base);
}

/*
 * Extract disk label
 */
static char *GetDiskLabel(dk_vtoc)
    DKvtoc		       *dk_vtoc;
{
    register char	       *cp;
    char		       *label;

    if (!dk_vtoc)
	return((char *)NULL);

    label = strdup(dk_vtoc->v_asciilabel);

    /*
     * The label normally has geometry information in it we don't want
     * to see, so we trim out anything starting with " cyl".
     */
    for (cp = label; cp && *cp; ++cp)
	if (*cp == ' ' && strncasecmp(cp, " cyl", 4) == 0)
	    *cp = C_NULL;

    return(label);
}

/*
 * Get the name of the controller for a disk.
 */
static char *GetDkCtlrName(dk_cinfo)
    DKcinfo 		       *dk_cinfo;
{
    if (!dk_cinfo)
	return((char *) NULL);

    (void) sprintf(Buff, "%s%d", dk_cinfo->dki_cname, dk_cinfo->dki_cnum);

    return(strdup(Buff));
}

/*
 * Get the disk controller model name from a disk.
 */
static char *GetDkCtlrModel(dk_cinfo)
    DKcinfo	 	       *dk_cinfo;
{
    struct dkctlrtab 	       *pct;

    if (!dk_cinfo)
	return((char *) NULL);

    if (!(pct = GetDkCtlrTab(dk_cinfo->dki_ctype)))
	return(NULL);

    return(pct->ct_model);
}

/*
 * Get a disk controller device from disk info.
 */
static DEVICE *GetDkCtlrDevice(DevData, dk_cinfo)
    DEVDATA 		       *DevData;
    DKcinfo	 	       *dk_cinfo;
{
    DEVICE 		       *MkMasterFromDevData();
    DEVICE 		       *dkctlr;

    if (!dk_cinfo)
	return((DEVICE *) NULL);
	
    if ((dkctlr = NewDevice(NULL)) == NULL)
	return((DEVICE *) NULL);

    BZERO((char *) dkctlr, sizeof(*dkctlr));

    dkctlr->dv_type = DT_DISKCTLR;

    /*
     * Get name of controller from devdata if available
     */
    if (DevData && DevData->dd_ctlrname)
	dkctlr = MkMasterFromDevData(DevData);

    if (!dkctlr->dv_name) {
	dkctlr->dv_name = GetDkCtlrName(dk_cinfo);
	dkctlr->dv_unit = dk_cinfo->dki_cnum;
    }
    dkctlr->dv_addr = dk_cinfo->dki_addr;
    dkctlr->dv_prio = dk_cinfo->dki_prio;
    dkctlr->dv_vec = dk_cinfo->dki_vec;
    dkctlr->dv_model = GetDkCtlrModel(dk_cinfo);

    return(dkctlr);
}

/*
 * Convert all we've learned about a disk to a DEVICE.
 */
static DEVICE *DKtoDiskDrive(devname, DevData,
			     dk_vtoc, dk_cinfo, dk_geom, dk_type)
    char 		       *devname;
    DEVDATA 		       *DevData;
    DKvtoc		       *dk_vtoc;
    DKcinfo	 	       *dk_cinfo;
    DKgeom 		       *dk_geom;
    DKtype 		       *dk_type;
{
    DEVICE 		       *Device, *dkctlr;
    DISKDRIVE 		       *diskdrive;
    int				listcnt = 0;

    if ((Device = NewDevice(NULL)) == NULL) {
	Error("Cannot create new device entry.");
	return((DEVICE *) NULL);
    }

    if ((dkctlr = NewDevice(NULL)) == NULL) {
	Error("Cannot create new dkctlr device entry.");
	return((DEVICE *) NULL);
    }

    if ((diskdrive = NewDiskDrive(NULL)) == NULL) {
	Error("Cannot create new diskdrive entry.");
	return((DEVICE *) NULL);
    }

    (void) sprintf(Buff, "%s%d", DevData->dd_devname, DevData->dd_devunit);
    Device->dv_aname 		= strdup(Buff);
    Device->dv_name 		= devname;
    Device->dv_type 		= DT_DISKDRIVE;
    Device->dv_nodeid 		= DevData->dd_nodeid;

    /*
     * Only read partition info we we're going to print it later.
     */
    if (VL_ALL)
	diskdrive->dd_part 	= GetDiskPart(devname, dk_vtoc);
    diskdrive->dd_label 	= GetDiskLabel(dk_vtoc);
    Device->dv_model 		= diskdrive->dd_label;

    if (dk_cinfo) {
	diskdrive->dd_unit 	= dk_cinfo->dki_unit;
	diskdrive->dd_slave 	= dk_cinfo->dki_slave;;
    }
    if (dk_geom) {
	diskdrive->dd_dcyl 	= dk_geom->dkg_ncyl;
	diskdrive->dd_pcyl 	= dk_geom->dkg_pcyl;
	diskdrive->dd_acyl 	= dk_geom->dkg_acyl;
	diskdrive->dd_heads 	= dk_geom->dkg_nhead;
	diskdrive->dd_sect 	= dk_geom->dkg_nsect;
	diskdrive->dd_apc 	= dk_geom->dkg_apc;
	diskdrive->dd_rpm 	= dk_geom->dkg_rpm;
	diskdrive->dd_intrlv 	= dk_geom->dkg_intrlv;
    }
    if (dk_type) {
	diskdrive->dd_psect 	= dk_type->hdkt_hsect;
	diskdrive->dd_promrev 	= dk_type->hdkt_promrev;
    }
    diskdrive->dd_secsize 	= dk_vtoc->v_sectorsz;

    Device->dv_desclist 	= (char **)xcalloc(MAXDESCLIST,sizeof(char *));
    Device->dv_desclist[listcnt++] 	= GetDiskCapacity(diskdrive);
    if (dk_vtoc->v_volume[0]) {
	(void) sprintf(Buff, "Volume Name is \"%s\"", dk_vtoc->v_volume);
	Device->dv_desclist[listcnt++] 	= strdup(Buff);
    }
    Device->dv_desclist[listcnt] 	= NULL;

    dkctlr 			= GetDkCtlrDevice(DevData, dk_cinfo);

    Device->dv_devspec 		= (caddr_t *) diskdrive;
    Device->dv_master 		= dkctlr;

    return(Device);
}

/*
 * Probe a disk drive
 */
extern DEVICE *ProbeDiskDrive(diskname, DevData, DevDataTab)
    /*ARGSUSED*/
    char 		       *diskname;
    DEVDATA 		       *DevData;
    DEVDATATAB 	     	       *DevDataTab;
{
    DEVICE 		       *diskdevice;
    DKcinfo 		       *dk_cinfo;
    DKgeom 		       *dk_geom;
    DKtype 		       *dk_type;
    DKvtoc 		       *dk_vtoc;
    char		       *devfile;
    char		       *devname;
    char		       *cp;
    int				fd;

    if (!diskname)
	return((DEVICE *)NULL);

    if (!(devfile = GetDeviceFile(DevData, _PATH_DEV_RDSK))) {
	if (Debug) Error("Cannot find disk device for `%s'.", diskname);
	return((DEVICE *)NULL);
    }

    /* Use basename of device file for virtual disk device name */
    devname = strdup(devfile);
    if (cp = strrchr(devname, '/'))
	devname = ++cp;
    /* Zap slice part of name */
    if (cp = strrchr(devname, 's'))
	*cp = C_NULL;

    if ((fd = open(devfile, O_RDONLY)) < 0) {
	if (Debug) Error("%s: Cannot open for reading: %s.", devfile, SYSERR);
	/*
	 * If we know for sure this drive is present and we
	 * know something about it, then create a minimal device.
	 */
	if ((DevDataTab->ddt_model || DevDataTab->ddt_desc) &&
	    FLAGS_ON(DevData->dd_flags, DD_IS_ALIVE)) {
	    Device = NewDevice((DEVICE *) NULL);
	    Device->dv_name = strdup(devname);
	    Device->dv_aname = strdup(diskname);
	    Device->dv_unit = DevData->dd_devunit;
	    Device->dv_master = MkMasterFromDevData(DevData);
	    Device->dv_type = DT_DISKDRIVE;
	    Device->dv_nodeid = DevData->dd_nodeid;
	    Device->dv_model = DevDataTab->ddt_model;
	    Device->dv_desc = DevDataTab->ddt_desc;
	    return(Device);
	} else
	    return((DEVICE *) NULL);
    }

    if ((dk_vtoc = GETvtoc(fd, devfile)) == NULL)
	if (Debug) Error("%s: get vtoc failed.", devfile);
    if ((dk_cinfo = GETdk_cinfo(fd, devfile)) == NULL)
	if (Debug) Error("%s: get dk_cinfo failed.", devfile);
    if ((dk_geom = GETdk_geom(fd, devfile)) == NULL)
	if (Debug) Error("%s: get dk_geom failed.", devfile);
    if ((dk_type = GETdk_type(fd, devfile)) == NULL)
	if (Debug) Error("%s: no dk_type info available.", devfile);

    close(fd);

    if (!(diskdevice = DKtoDiskDrive(devname, DevData,
				     dk_vtoc, dk_cinfo, dk_geom, dk_type))) {
	Error("%s: Cannot convert diskdrive information.", devname);
	return((DEVICE *) NULL);
    }

    return(diskdevice);
}

/*
 * Get hostid
 */
long gethostid()
{
    char 			buff[BUFSIZ];

    sysinfo(SI_HW_SERIAL, buff, sizeof(buff));

    return(atol(buff));
}

/*
 * Get system page size
 */
int getpagesize()
{
    long			siz;

    if ((siz = sysconf(_SC_PAGESIZE)) == -1) {
	Error("Cannot get pagesize from sysconf(): %s", SYSERR);
	return(0);
    }

    return((int) siz);
}

#if	defined(HAVE_SUNROMVEC)
#include <sys/comvec.h>
#endif	/* HAVE_SUNROMVEC */

/*
 * Get ROM Version number
 */
extern char *GetRomVer()
{
    static char			RomRev[32];
#if	defined(HAVE_SUNROMVEC)
    struct nlist	       *nlptr;
    static union sunromvec	Rom;
    register char	       *p;
    register char	       *Addr;
    kvm_t		       *kd;
    union sunromvec	       *romp;

    if (!(kd = KVMopen()))
	return((char *) NULL);

    if ((nlptr = KVMnlist(kd, RomVecSYM, (struct nlist *)NULL)) == NULL)
	return((char *) NULL);

    if (CheckNlist(nlptr))
	return((char *) NULL);

    /*
     * Read the kernel pointer to the sunromvec structure.
     */
    if (KVMread(kd, (u_long) nlptr->n_value, (char *) &romp, 
		sizeof(romp))) {
	if (Debug) Error("Cannot read sunromvec pointer from kernel.");
	return((char *) NULL);
    }

    /*
     * Read sunromvec from the kernel
     */
    if (KVMread(kd, (u_long) romp, (char *) &Rom, sizeof(struct sunromvec))) {
	if (Debug) Error("Cannot read sunromvec from kernel.");
	return((char *) NULL);
    }

    if (Rom.sunmon.v_mon_id) {
	/*
	 * Read the version string from the address indicated by v_mon_id.
	 * Read 1 byte at a time until '\0' is encountered.
	 */
	p = RomRev;
	Addr = Rom.sunmon.v_mon_id;
	do {
	    if (KVMread(kd, (u_long) Addr++, p, 1))
		break;
	} while (p < &RomRev[sizeof(RomRev)-1] && *p++);
	*p = C_NULL;
    }

    if (Rom.obp.op_mon_id && !RomRev[0]) {
	/*
	 * XXX Hardcoded values
	 */
	(void) sprintf(RomRev, "%d.%d", 
		       Rom.obp.op_mon_id >> 16, Rom.obp.op_mon_id & 0xFFFF);
    }

    KVMclose(kd);

#endif	/* HAVE_SUNROMVEC */

    return((RomRev[0]) ? RomRev : (char *) NULL);
}

/*
 * Get amount of physical memory
 */
extern int GetMemoryAmount()
{
    long			Amount;

    /*
     * sysmem() is an undocumented function that returns the
     * amount of system (physical) memory in bytes.  It's in
     * /usr/lib/libadm.a.  Discovered by reference in Sun
     * PatchID 101050-01.
     */
    Amount = sysmem();
    if (Amount < 0) {
	if (Debug) Error("sysmem failed: %s", SYSERR);
	return(-1);
    }

    return(DivRndUp((u_long) Amount, MBYTES));
}
